# -*- coding: utf-8 -*-
from .repetition import repetition_experiences

# TODO: sortir DialogRepetition de repetition_experiences.py -> __init__.py
# pour que repetition_experiences.py n'importe plus pyQt (risque de ralentir l'import pour ptyx !)

def parse_text(text):
    """Describe drawing.

    Coordinates are expressed as percentages of picture dimensions.
    Return a dict containing 2 lists:
     * one for points: [(text, x, y), ...]
     * one for lines: [(text, xy1, xy2), ...]
    """
    lines = [l for l in text.split("\n") if l.strip()]

    if not lines:
        return {'points': [], 'lines': []}

    if lines[0].startswith("|"):
        legende = lines.pop(0)
        # On réserve de la place pour la légence
        ymax = 90
    else:
        legende = None
        ymax = 100

    if not lines:
        return {'points': [], 'lines': []}

    # Répétition d'expériences aléatoires indépendantes
    #
    # Ex:
    # >> A:  1/3
    # >> &A: 2/3
    #
    # équivaut à:
    #
    # > A_1:   1/3
    # >> A_2:  1/3
    # >> &A_2: 2/3
    # > &A_1:  2/3
    # >> A_2:  1/3
    # >> &A_2: 2/3

    if lines[0].startswith('>>'):
        profondeur = len(lines[0]) - len(lines[0].lstrip('>'))
        evenements = {}
        for line in lines:
            assert line.startswith(profondeur*'>')
            nom, proba = line[profondeur:].split(':')
            evenements[nom.strip()] = proba.strip()
        lines = repetition_experiences(_profondeur=profondeur, _numeroter=True, **evenements).split('\n')

    # La 1re ligne doit correspondre au 1er embranchement, c'est-à-dire à l'univers (Omega en général).
    # À défaut, on met un ligne vide (pas de nom au 1er embranchement donc).
    if lines[0].startswith(">"):
        lines = [""] + lines

    depths = [len(line) - len(line.lstrip(">")) for line in lines]

    nbr_colonnes = 1 + max(depths, default=0)

    # Nbr de lignes pour chaque colonne.
    nbr_lignes = []

    for i in xrange(nbr_colonnes):
        nbr_lignes.append(depths.count(i))

    #intersection, union : \cap \cup

    # Interprétation des lines sous forme de listes de listes
    # >A
    # >>B
    # >>C
    # >D
    # >>E
    # >>>F
    # >>>G
    # >>H
    # >E
    # devient :
    # [{'liste': [{'liste': [{'liste': [], 'texte': u'B'}, {'liste': [], 'texte': u'C'}], 'texte': u'A'}, {'liste': [{'liste': [{'liste': [], 'texte': u'F'}, {'liste': [], 'texte': u'G'}], 'texte': u'E'}, {'liste': [], 'texte': u'H'}], 'texte': u'D'}, {'liste': [], 'texte': u'E'}], 'texte': ''}]

    arbre = []
    ligne_precedente = [-1 for i in xrange(nbr_colonnes)] # numéro de ligne atteint pour chaque colonne
    for line in lines:
        colonne = len(line) - len(line.lstrip(">"))
        branche = arbre
        for i in xrange(colonne): # on se déplace de branche en branche ;)
            branche = branche[ligne_precedente[i]]["liste"]
        branche.append({"texte": line.lstrip(">"), "liste": []})
        ligne_precedente[colonne] += 1
        for i in xrange(colonne + 1, nbr_colonnes):
            ligne_precedente[i] = -1
    print arbre

    # on parcourt l'arbre pour compter le nombre de ramifications

    def compter_ramifications(branche):
        if len(branche["liste"]) > 0:
            return sum(compter_ramifications(tige) for tige in branche["liste"])
        else:
            return 1

    ramifications = sum(compter_ramifications(branche) for branche in arbre)
    if param.debug:
        print("Nombre de ramifications: " + str(ramifications))

    def formater_texte(texte):
        if texte:
            if texte.startswith("&"):
                texte = r"\overline{" + texte[1:] + "}"
            texte = texte.replace("&", r"\overline ")
            # P(\overline A) -> P(\overline{A})
            texte = re.sub("(\\\\overline)[ ]+(%s)" % VAR, lambda m:'%s{%s}' % (m.group(1), m.group(2)), texte)
            texte = "$" + texte + "$" if texte[0] != '$' else texte
            if param.latex:
                texte = "$" + texte + "$" # passage en mode "display" de LaTeX
            texte = texte.replace(" inter ", r"\  \cap \ ").replace(" union ", r"\  \cup \ ").replace("Omega", r"\Omega").replace("omega", r"\Omega")
            if param.latex: # on remplace les fractions. Ex: "1/15" -> "\frac{1}{15}"
                texte = re.sub("[0-9]+/[0-9]+",lambda s:"\\frac{" + s.group().replace("/", "}{") + "}", texte)
        return texte


    def creer_point(x, y, texte):
        texte = formater_texte(texte)
        M = Point(x, y, style = "o", couleur = "w", taille = 0)
        M.label(texte)
        M.etiquette.style(_rayon_=0, niveau=15, alignement_vertical="center",
                          alignement_horizontal="center", fond=True,
                          couleur_fond="w")
        return M


    def creer_segment(point1, point2, texte):
        texte = formater_texte(texte)
        s = Segment(point1, point2)
        s.label(texte)
        style = {'_rayon_': 0, 'niveau': 15}
        placement = self.param('placement_probabilites')
        if placement == 'dessus':
            style.update(alignement_vertical='center', couleur_fond='w',
                         fond=True, alignement_horizontal='center',
                         angle=0)
        elif placement == 'longe':
            style.update(alignement_vertical='auto', fond=False,
                         alignement_horizontal='center', angle='auto')
        elif placement == 'decale':
            style.update(alignement_vertical='auto', fond=False,
                         alignement_horizontal='right', angle=0)
        else:
            print(u"Placement: mode '%s' non reconnu." % placement)
        s.etiquette.style(**style)
        return s

    ramification = [0] # astuce pour avoir un "objet modifiable" (a mutable object)
    def parcourir_branche(branche, n, ramification = ramification):
        txt = branche["texte"].rsplit(":", 1)
        if len(txt) == 2:
            txt_pt, txt_segm = txt
        else:
            txt_pt = txt[0]
            txt_segm = ""
        if len(branche["liste"]) > 0:
            l = []
            for tige in branche["liste"]:
                l.append(parcourir_branche(tige, n + 1))

            M = creer_point(n/(nbr_colonnes - 1), .5*(l[0][0].ordonnee+l[-1][0].ordonnee), txt_pt)
            self.feuille_actuelle.objets.add(M)
            for point, txt_s in l:
                s = creer_segment(M, point, txt_s)
                self.feuille_actuelle.objets.add(s)
            return M, txt_segm
        else:
            if ramifications > 1:
                y = 1 - ramification[0]/(ramifications - 1)
            else:
                y = .5
            M = creer_point(n/(nbr_colonnes - 1), y, txt_pt)
            ramification[0] += 1
            self.feuille_actuelle.objets.add(M)
            return M, txt_segm

    for branche in arbre:
        parcourir_branche(branche, 0)

    if legende is not None:
        decalage = -0.5
        for i in legende:
            if i != "|":
                break
            decalage += .5
        legende = legende.strip("|").split("|")
        for n in xrange(len(legende)):
            t = Texte(legende[n], (n + decalage)/(nbr_colonnes - 1), 1.1)
            self.feuille_actuelle.objets.add(t)

    self.feuille_actuelle.interprete.commande_executee()



# -*- coding: utf-8 -*-

##--------------------------------------#######
#           Mathlib 2 (sympy powered)         #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

## Cette librairie contient les fonctions de haut niveau non inclues dans sympy.

import math

from sympy import Rational, sqrt, S
from sympy.stats import Normal, Binomial, P
import sympy.stats






def _convertir_en_frequences(frequences, serie):
    if frequences is None:
        n = len(serie)
        return n*[Rational(1, n)]
    else:
        total = sum(frequences)
        if total != 1:
            return [val/total for val in frequences]
        return frequences

def moyenne(serie, coeffs = None):
    "Calcule l'espérance de la série des (xi, fi)."
    frequences = _convertir_en_frequences(coeffs, serie)
    return sum(xi*fi for xi, fi in zip(serie, frequences))

def variance(serie, coeffs = None):
    "Calcule la variance de la serie des (xi, fi)"
    frequences = _convertir_en_frequences(coeffs, serie)
    M = moyenne(serie, frequences)
    return sum(fi*(xi - M)**2 for xi, fi in zip(serie, frequences))

def ecart_type(serie, coeffs = None):
    "Retourne l'écart-type de la série des (xi, fi)."
    return sqrt(variance(serie, coeffs))

def covariance(serie1, serie2, coeffs = None):
    "Retourne la covariance des deux séries."
    assert len(serie1) == len(serie2), "Les deux séries doivent avoir le même nombre de valeurs."
    frequences = _convertir_en_frequences(coeffs, serie1)
    x_ = moyenne(serie1, frequences)
    y_ = moyenne(serie2, frequences)
    return sum(fi*(xi - x_)*(yi - y_) for xi, yi, fi in zip(serie1, serie2, frequences))

def linreg(serie1, serie2, coeffs = None):
    """Droite de régression par la méthode des moindres carrés.

    Retourne les coefficients a et b de l'équation y=ax+b
     de la droite de régression par la méthode des moindres carrés.

     >>> from wxgeometrie.mathlib.custom_functions import linreg
     >>> linreg((85.6,84.5,81,80.2,72.8,71.2,73,48.1),(78.7,77.6,75.2,71.1,67.7,66.3,59.1,46.8))
     (0.849191825268073, 4.50524942626518)
     """
    a = covariance(serie1, serie2)/variance(serie1)
    b = moyenne(serie2) - a*moyenne(serie1)
    return a, b




def proba(relation):
    if relation is True:
        return S.One
    elif relation is False:
        return S.Zero
    else:
        return P(relation)


def inv_normal(p):
    """
    Lower tail quantile for standard normal distribution function.

    This function returns an approximation of the inverse cumulative
    standard normal distribution function.  I.e., given P, it returns
    an approximation to the X satisfying P = Pr{Z <= X} where Z is a
    random variable from the standard normal distribution.

    The algorithm uses a minimax approximation by rational functions
    and the result has a relative error whose absolute value is less
    than 1.15e-9.

    Author:      Peter John Acklam
    Time-stamp:  2000-07-19 18:26:14
    E-mail:      pjacklam@online.no
    WWW URL:     http://home.online.no/~pjacklam
    """

    if p <= 0 or p >= 1:
        # The original perl code exits here, we'll throw an exception instead
        raise ValueError("Argument to inv_normal %f must be in open interval (0,1)" % p)

    # Coefficients in rational approximations.
    a = (-3.969683028665376e+01,  2.209460984245205e+02, \
         -2.759285104469687e+02,  1.383577518672690e+02, \
         -3.066479806614716e+01,  2.506628277459239e+00)
    b = (-5.447609879822406e+01,  1.615858368580409e+02, \
         -1.556989798598866e+02,  6.680131188771972e+01, \
         -1.328068155288572e+01 )
    c = (-7.784894002430293e-03, -3.223964580411365e-01, \
         -2.400758277161838e+00, -2.549732539343734e+00, \
          4.374664141464968e+00,  2.938163982698783e+00)
    d = (7.784695709041462e-03,  3.224671290700398e-01, \
          2.445134137142996e+00,  3.754408661907416e+00)

    # Define break-points.
    plow  = 0.02425
    phigh = 1 - plow

    # Rational approximation for lower region:
    if p < plow:
       q  = math.sqrt(-2*math.log(p))
       return (((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) / \
               ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1)

    # Rational approximation for upper region:
    if phigh < p:
       q  = math.sqrt(-2*math.log(1-p))
       return -(((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) / \
                ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1)

    # Rational approximation for central region:
    q = p - 0.5
    r = q*q
    return (((((a[0]*r+a[1])*r+a[2])*r+a[3])*r+a[4])*r+a[5])*q / \
           (((((b[0]*r+b[1])*r+b[2])*r+b[3])*r+b[4])*r+1)


def fluctu(probabilite, taille=1000, seuil=None):
    """Intervalle de fluctuation.

    Retourne l'intervalle de fluctuation correspondant à un échantillon de taille
    `taille`, lorsque la probabilité est `probabilite`.

    Par défaut, lorsque le seuil n'est pas fixé, l'approximation
    [p - 1.96*sqrt(p(1-p)/n), p + 1.96*sqrt(p(1-p)/n)] est utilisée,
    pour un seuil environ égal à 95%.
    """
    if seuil is None:
        # Approximation usuelle pour un seuil à 95%
        a = 1.96
    else:
        # P(-a < X < a) == seuil <=> P(X < a) == (seuil + 1)/2
        a = inv_normal((seuil + 1)/2)
    delta = a*math.sqrt(probabilite*(1 - probabilite)/taille)
    return probabilite - delta, probabilite + delta


def confiance(frequence, taille=1000, seuil=None):
    """Intervalle de confiance au seuil de 95%.

    Retourne l'intervalle de confiance correspondant à un échantillon de taille
    `taille`, lorsque la fréquence observée sur l'échantillon est `frequence`.

    Par défaut, si seuil n'est pas fixé, l'approximation utilisée est
    [f - 1/sqrt(n), f + 1/sqrt(n)], pour un seuil environ égal à 95%.
    """
    if seuil is not None:
        return fluctu(frequence, taille, seuil)
    delta = 1/math.sqrt(taille)
    return frequence - delta, frequence + delta


def normal(a, b, mu=0, sigma=1):
    """Retourne P(a < X < b), où X suit la loi normale N(mu, sigma²).
    """
    X = Normal('X', mu, sigma)
    return (proba(X <= b) - proba(X < a)).evalf()


def binomial(a, b, n, p):
    """Retourne P(a <= X <= b), où X suit la loi binomiale B(n, p).

    ..note:: Taper binomial(a, a, n, p) pour calculer P(X = a).
    """
    X = Binomial('X', n, p)
    return (proba(X <= b) - proba(X < a)).evalf()


def va(loi, *parametres):
    # [key for key, val in sympy.stats.__dict__.items()
    #                if isinstance(val, types.FunctionType)
    #                   and len(key) > 1 and key[0].isupper()]
    loi = loi.capitalize()
    synonymes = {'Binomiale': 'Binomial',  'Hypergeometrique': 'Hypergeometric',
                 'Uniforme': 'Uniform', 'De': 'Die', 'Piece': 'Coin',
                 'Triangulaire': 'Triangular', 'Normale': 'Normal',
                }
    loi = synonymes.get(loi, loi)
    try:
        return sympy.stats.__dict__[loi]('X', *parametres)
    except KeyError:
        raise KeyError('Loi "%s" inconnue !' % loi)



# -*- coding: utf-8 -*-

#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


# Ce module sert à tester que toutes les dépendances sont satisfaites,
# et à y remédier dans le cas contraire (si possible).

import sys, imp, platform, os, shutil, subprocess

# ------------------------------------------------------------------------------
# CONFIGURATION
# ------------------------------------------------------------------------------

from .version import NOMPROG

# Les valeurs sont les noms des paquets sous Debian/Ubuntu.
dependances = {'PyQt5': 'python3-pyqt5',
               'matplotlib': 'python3-matplotlib',
               'scipy': 'python3-scipy',
               'numpy': 'python3-numpy',
               'mpmath': 'python3-mpmath',
               'PyQt5.Qsci': 'python3-pyqt5.qsci',
               'PyQt5.QtSvg': 'python3-pyqt5.qtsvg',
               }


# Most of the time, the name of the pip package is just the module name.
# However, those are exceptions:
pip_packages = {'PyQt5.Qsci': 'Qscintilla',
               'PyQt5.QtSvg': '', # already included in PyQt5
               }


python_version_min = (3, 5)
python_version_max = (3, 99)

plateforme = platform.system() #'Windows' ou 'Linux' par exemple.

# Paramètres de matplotlib
# Utiliser une installation LaTeX existante (meilleur rendu mais très lent !)
latex = False
latex_unicode = True
moteur_de_rendu = 'Qt5Agg'

# ------------------------------------------------------------------------------


def tester_dependances():
    if hasattr(sys, 'frozen'):
        # Ne pas faire ces tests avec py2exe/py2app/CxFreeze (non seulement inutiles, mais en plus ils échouent).
        return
    # Make sure I have the right Python version.
    version_python_supportee = True
    if sys.version_info[:2] < python_version_min:
        print(" ** Erreur fatale **")
        print(NOMPROG + " nécessite Python %d.%d au minimum." % python_version_min)
        version_python_supportee = False
    elif sys.version_info[:2] > python_version_max:
        print(" ** Erreur fatale **")
        print(NOMPROG + " supporte Python %d.%d au maximum." % python_version_max)
        version_python_supportee = False

    if not version_python_supportee:
        print("Python %d.%d détecté." % sys.version_info[:2])
        sys.exit(-1)

    # Test for dependencies:
    modules_manquants = []
    for module in dependances:
        module_name = module
        try:
            # imp.find_module() doesn't support submodules.
            path = None
            while '.' in module:
                module, submodule = module.split('.', 1)
                f, filename, description = imp.find_module(module, path)
                module = submodule
                path = [filename]
            imp.find_module(module, path)
        except ImportError:
            modules_manquants.append(module_name)



    if modules_manquants:
        try:
            if not sys.stdin.isatty():
                raise ImportError
            if input(f"Certaines librairies manquent : {modules_manquants}.\n"
                      "Voulez-vous les installer automatiquement ? (o/N)"
                     ).lower() not in ('o', 'oui', 'yes', 'y'):
                 raise ImportError
            # Try to install missing modules using pip.
            pip_modules = [pip_packages.get(module, module) for module in modules_manquants]
            try:
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--user'] + pip_modules)
            except subprocess.CalledProcessError:
                # Maybe we are in a virtual environment, so --user is meaningless ?
                try:
                    subprocess.check_call([sys.executable, '-m', 'pip', 'install'] + pip_modules)
                except subprocess.CalledProcessError:
                    print("L'installation automatique des modules manquants a échouée.")
                    raise ImportError
            print('Des modules indispensables ont été installés.')
            print('Redémarrage de Géophar...')
            os.execv(sys.executable, [sys.executable] + sys.argv)
            sys.exit()
        except ImportError:
            if plateforme == 'Linux' and shutil.which('apt'):
                paquets = ' '.join(dependances[m] for m in modules_manquants)
                print(f"""
[\033[91mERREUR\033[39m] Certains paquets sont manquants : {paquets}.

Votre distribution semble être dérivée de Debian ou d'Ubuntu,
il est possible de remédier à cette situation en lançant la commande :
----------------------------------------------------------------------
\033[1m\
sudo apt update; sudo apt install {paquets}\
\033[0m
----------------------------------------------------------------------
Vous pourrez ensuite relancer Géophar.
""")
        else:
            print('** Erreur fatale **\nLes modules suivants sont introuvables !')
            print('MODULE(S) MANQUANT(S): %s.' % ', '.join(modules_manquants))
        sys.exit(-1)


def configurer_dependances():
    import PyQt5

    # ---------------------------
    # Configuration de Matplotlib
    # ---------------------------
    try:
        # Cx_Freeze version needs this (so that matplotlib can found pyparsing).
        import pyparsing
    except ImportError:
        pass
    import matplotlib
    matplotlib.rcParams['backend'] = 'Qt5Agg'
    try:
        # For old versions of matplotlib (version < 3.0)
        if matplotlib.__version__.split(".")[0] < '3':
            matplotlib.rcParams['backend.qt5'] ='PyQt5'
            matplotlib.rcParams["text.latex.unicode"] = latex_unicode
    except KeyError:
        pass
    matplotlib.use(moteur_de_rendu)
    matplotlib.rcParams['text.usetex'] = latex

    # A changer *avant* d'importer pylab ?
    matplotlib.rcParams['font.family'] ='serif'
    #matplotlib.rcParams['font.sans-serif'] ='STIXGeneral'
    matplotlib.rcParams['font.serif'] ='STIXGeneral'
    #matplotlib.rcParams['font.monospace'] ='STIXGeneral'
    matplotlib.rcParams['mathtext.fontset'] ='stix'




# -*- coding: utf-8 -*-

##---------------------------------------#######
#                  Backends                    #
##---------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2013  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


# Ce module contient les différents backends, qui servent à exporter la feuille.

from .backend_tikz import BackendTikz

# Le dictionnaire suivant permet à la feuille de connaître les différents
# backends disponibles.

backends_dict = {'tikz': BackendTikz(),
                }
